#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <fcntl.h>
#include <signal.h>
#ifdef TURBOC
#include <dos.h>
#include <alloc.h>
#include <graphics.h>
#else
#include <console.h>
#endif
#include "wirefram.h"

#undef min
#undef max
#define min(x,y) (x<y?x:y)
#define max(x,y) (x>y?x:y)

#define Rotate_Option0 1
#define Rotate_Option1 2
#define Rotate_Option2 4
#define Rotate_Option3 8
#define Translate_Option1 16
#define Translate_Option2 32

/* extern int _fmode = O_BINARY; */
#ifdef TURBOC
extern unsigned _stklen = 10000;
#endif

/* Graphics variables */
#ifdef TURBOC
extern struct viewporttype actual_view;
#else
extern Rect actual_view;
#define BLUE 55
#endif
extern double offsetx, offsety; /* The actual center of the screen in pixels */
extern int window_width, window_height;

FILE *outfile;
char title[256];
int Draw_Option; /* Which line draw routine? */
unsigned long int PATTERN;
int LINEWIDTH;
int All_Flag = 0;
int Shade_Flag = 0;
int Normals_Flag = 0;
int Hidden_Flag = 0;
int Draw_Axes_Flag = 0;
int Show_Orientation_Flag = 1;
int Options = Rotate_Option2;
DBL VTemp;
OBJECT *Light_Sources = NULL;
OBJECT *All_Objects = NULL;
OBJECT *Current_Object = NULL;
OBJECT *Current_Display_Object = NULL;
VECTOR TempVec;
unsigned long zbufsiz = 0;


TEXTURE RGB_Red     = {{ 1.0, 0.0, 0.0 }, 1.0};
TEXTURE RGB_Magenta = {{ 1.0, 0.0, 1.0 }, 1.0};
TEXTURE RGB_Blue    = {{ 0.0, 0.0, 1.0 }, 1.0};
TEXTURE RGB_Cyan    = {{ 0.0, 1.0, 1.0 }, 1.0};
TEXTURE RGB_Green   = {{ 0.0, 1.0, 0.0 }, 1.0};
TEXTURE RGB_Yellow  = {{ 1.0, 1.0, 0.0 }, 1.0};
TEXTURE RGB_Black   = {{ 0.0, 0.0, 0.0 }, 1.0};
TEXTURE RGB_White   = {{ 1.0, 1.0, 1.0 }, 1.0};

Depth_Map_Type Depth_Map[64];
Screen_Map_Type Screen_Map[64];
unsigned long Total_Map_Lines, Total_Map_Buffers, Total_Elements;
unsigned long First_Map_Line, Last_Map_Line;
unsigned long Total_Lines_Per_Buffer, Total_Elements_Per_Buffer;

/* Forward function declarations */
DBL degrees_to_radians(DBL degrees);
void render_current_object();
void render_objects();

VIEWPOINT *Eye; /* The default viewpoint */
TRANSFORMATION WS;

#ifdef TURBOC
int _Cdecl
matherr(e)
   struct exception *e;
{
   /* Since we are just making pictures, not keeping nuclear power under
      control - it really isn't important if there is a minor math problem.
      This routine traps and ignores them.  Note: the most common one is
      a DOMAIN error coming out of "acos". */
   switch (e->type) {
      case DOMAIN   : printf("DOMAIN error in '%s'\n", e->name); break;
      case SING     : printf("SING   error in '%s'\n", e->name); break;
      case OVERFLOW : printf("OVERFLOW error in '%s'\n", e->name); break;
      case UNDERFLOW: printf("UNDERFLOW error in '%s'\n", e->name); break;
      case TLOSS    : printf("TLOSS error in '%s'\n", e->name); break;
      case PLOSS    : printf("PLOSS error in '%s'\n", e->name); break;
      case EDOM     : printf("EDOM error in '%s'\n", e->name); break;
      case ERANGE   : printf("ERANGE error in '%s'\n", e->name); break;
      default       : printf("Unknown math error in '%s'\n",e->name);break;
      }
   return (1);  /* Indicate the math error was corrected... */
}
#endif

void
#ifdef TURBOC
Signal_Catcher(int sig, int subcode)
#else
Signal_Catcher(int sig)
#endif
{
   switch (sig) {
   case SIGABRT: printf("SIGABRT caught\n"); break;
   case SIGFPE:
      printf("SIGFPE caught\n");
#ifdef TURBOC
      switch(subcode) {
      case FPE_INTOVFLOW:
         printf("INTO executed with OF flag set\n");
         break;
      case FPE_INTDIV0:
         printf("Integer divide by 0\n");
         break;
      case FPE_INVALID:
         printf("Invalid operation\n");
         break;
      case FPE_ZERODIVIDE:
         printf("Division by zero\n");
         break;
      case FPE_OVERFLOW:
         printf("Numeric overflow\n");
         break;
      case FPE_UNDERFLOW:
         printf("Numeric underflow\n");
         break;
      case FPE_INEXACT:
         printf("Precision error\n");
         break;
      case FPE_EXPLICITGEN:
         printf("User program executed raise(SIGFPE)\n");
         break;
      default:
         printf("FPU subcode: %d unknown\n", subcode);
      }
#endif
      /* Put a break on top of a return statement here and trace back
         through the stack to see why bad floating intstructions occurred. */
      break;
   case SIGILL: printf("SIGILL caught\n"); break;
   case SIGINT: printf("SIGINT caught\n"); break;
   case SIGSEGV: printf("SIGSEGV caught\n"); break;
   case SIGTERM: printf("SIGTERM caught\n"); break;
   default: printf("Signal %d caught\n", sig); break;
   }
   exit(1);
}

void
init_signal_catcher()
{
   signal(SIGABRT, Signal_Catcher);
   signal(SIGFPE, Signal_Catcher);
   signal(SIGILL, Signal_Catcher);
   signal(SIGINT, Signal_Catcher);
   signal(SIGSEGV, Signal_Catcher);
   signal(SIGTERM, Signal_Catcher);
}

void
add_object(Object)
   OBJECT *Object;
{
   Object->Next_Object = All_Objects;
   All_Objects = Object;
}

void
add_light(Object)
   OBJECT *Object;
{
   Object->Next_Object = Light_Sources;
   Light_Sources = Object;
}

void
edit_lights()
{
   int i;
   OBJECT *Light;
   restorecrtmode();
   for (i=0,Light = Light_Sources;
        Light != NULL;
        i++, Light = Light->Next_Object) {
      printf("Edit light %d\n", i);
      edit_light(Light);
      }
   setgraphmode(getgraphmode());
   reset_pallette();
}

void
drawaxes()
{
   int i;
   VECTOR pt1[8], vd;
   DBL vx1[8],vy1[8];
   OBJECT *Temp_Light;

   /* Coordinate axes */
   VSet(pt1[0], -3.0,  0.0,  0.0);
   VSet(pt1[1],  3.0,  0.0,  0.0);
   VSet(pt1[2],  0.0, -3.0,  0.0);
   VSet(pt1[3],  0.0,  3.0,  0.0);
   VSet(pt1[4],  0.0,  0.0, -3.0);
   VSet(pt1[5],  0.0,  0.0,  3.0);
}

void
modify_object(Object)
   OBJECT *Object;
{
   get_orientation(Object);
   get_parameters(Object);
   get_unique_params(Object);
   if (Object->Object_Type != Viewpoint_Type)
      get_colors(Object);
}

void
delete_object(Object)
   OBJECT *Object;
{
   char ch;
   OBJECT *New_Object, *Temp_Object;
   int Object_Type = Object->Object_Type;

   if (Current_Object == NULL || All_Objects == NULL)
      return;
   else if (Object == All_Objects) {
      All_Objects = All_Objects->Next_Object;
      Current_Object = All_Objects;
      }
   else {
      Temp_Object = All_Objects;
      while (Temp_Object->Next_Object != Object)
         Temp_Object = Temp_Object->Next_Object;
      Temp_Object->Next_Object = Object->Next_Object;
      if (Object->Next_Object == NULL)
         Current_Object = All_Objects;
      else
         Current_Object = Object->Next_Object;
      }
   if (Object->Parameters != NULL)
      free(Object->Parameters);
   /* Free memory specific to polygons */
   if (Object_Type == Polygon_Type)
      free((void *)((POLYGON *)Object)->Polygon_Params->Vertices);
   /* When Bezier patch memory is dynamically allocated, put something
      in right here to free it. */
   free((void *)Object);
}

void
delete_light(Object)
   OBJECT *Object;
{
   OBJECT *Temp_Object;

   if (All_Objects == NULL)
      return;
   else if (Object == Light_Sources)
      Light_Sources = Light_Sources->Next_Object;
   else {
      Temp_Object = Light_Sources;
      while (Temp_Object->Next_Object != Object)
         Temp_Object = Temp_Object->Next_Object;
      Temp_Object->Next_Object = Object->Next_Object;
      if (Object->Next_Object == NULL)
         Current_Object = Light_Sources;
      else
         Current_Object = Object->Next_Object;
      }
   free((void *)Object);
}

void
create_new_object()
{
   char ch, text_line[256];
   OBJECT *New_Object = NULL;

   printf("1) Bezier      2) Ellipse  3) Hyperboloid\n");
   printf("4) Paraboloid  5) Plane    6) Torus\n");
   printf("7) Coil        8) Hyperbolic Paraboloid\n");
   printf("9) Polygon     0) Squishy  L) \n");
   printf("C) Cone\n");
   printf("A) Abort\n");
   printf("Enter Type number: ");

   l0:
   if (gets(text_line) == NULL)
      return;
   else
      ch = text_line[0];

   l1:
   if (ch != 'a' && ch != 'A') {
      switch (ch) {
      case 'c':
      case 'C':
         New_Object = create_object(Cone_Type);
         break;
      case 'l':
      case 'L':
         New_Object = create_object(Light_Source_Type);
         break;
      case '1':
         New_Object = create_object(Bezier_Type);
         break;
      case '2':
         New_Object = create_object(Ellipse_Type);
         break;
      case '3':
         New_Object = create_object(Hyperboloid_Type);
         break;
      case '4':
         New_Object = create_object(Paraboloid_Type);
         break;
      case '5':
         New_Object = create_object(Plane_Type);
         break;
      case '6':
         New_Object = create_object(Torus_Type);
         break;
      case '7':
         New_Object = create_object(Coil_Type);
         break;
      case '8':
         New_Object = create_object(Hyperbolic_Paraboloid_Type);
         break;
      case '9':
         New_Object = create_object(Polygon_Type);
         break;
      case '0':
         New_Object = create_object(Squishy_Type);
         break;
         }
      if (New_Object == NULL) {
         printf("Failed to create object\n");
         return;
         }
      else if (New_Object->Object_Type == Light_Source_Type) {
         edit_light(New_Object);
         add_light(New_Object);
         }
      else if (New_Object->Object_Type != Viewpoint_Type) {
         get_orientation(New_Object);
         get_parameters(New_Object);
         get_unique_params(New_Object);
	 get_colors(New_Object);
         if (New_Object != NULL) {
            add_object(New_Object);
            Current_Object = New_Object;
            }
         }
      }
}

void
prepare_DKB_output()
{
   FILE *dkb_file;
   char dkb_file_name[80];
   int fields;

   restorecrtmode();
   l0:
   printf("DKB output file (cr to abort)? ");
   if (gets(&dkb_file_name[0]) != NULL) {
      dkb_file = fopen(dkb_file_name, "wt");
      if (dkb_file == NULL) {
         printf("Bad file name: '%s'\n", dkb_file_name);
         goto l0;
         }
      }
   else
      return;
   write_DKB_information(dkb_file, Eye, Light_Sources, All_Objects);
   fclose(dkb_file);
   setgraphmode(getgraphmode());
   reset_pallette();
}

void
prepare_triangle_output(flag)
   int flag;
{
   FILE *file;
   char file_name[80];
   int fields;

   l0:
   printf("Triangle output file (cr to abort)? ");
   if (gets(&file_name[0]) != NULL) {
      file = fopen(file_name, "wt");
      if (file == NULL) {
         printf("Bad file name: '%s'\n", file_name);
         goto l0;
         }
      }
   else
      return;
   write_triangle_information(file, All_Objects, flag);
   fclose(file);
}

void
dump_wire_data()
{
   char file_name[32];
   OBJECT *Object;

   restorecrtmode();
   l0:
   printf("Wire output file (cr to abort)? ");
   if (gets(&file_name[0]) != NULL) {
      outfile = fopen(file_name, "wt");
      if (outfile == NULL) {
         printf("Bad file name: '%s'\n", file_name);
         goto l0;
         }
      }
   else
      return;

   wire_write_information((OBJECT *)Eye, outfile);
   for (Object = All_Objects; Object != NULL; Object = Object->Next_Object)
      wire_write_information(Object, outfile);
   for (Object = Light_Sources; Object != NULL; Object = Object->Next_Object) {
      wire_write_information(Object, outfile);
      }
   fclose(outfile);
   setgraphmode(getgraphmode());
   reset_pallette();
}

void
read_wire_data(wire_file)
   FILE *wire_file;
{
   OBJECT *Object;
   char text_line[80];
   int Object_Type;

   rewind(wire_file);
   while (!feof(wire_file)) {
      Object_Type = -1;
      fgets(&text_line[0], 255, wire_file);
      if (feof(wire_file)) return;
      sscanf(text_line, "%d", &Object_Type);
      if (Object_Type == Viewpoint_Type) {
         wire_read_information((OBJECT *)Eye, wire_file);
         }
      else if (Object_Type == Light_Source_Type) {
         Object = create_object(Object_Type);
         if (Object == NULL) return;
         wire_read_information(Object, wire_file);
         add_light(Object);
         }
      else {
         Object = create_object(Object_Type);
         if (Object == NULL) return;
         wire_read_information(Object, wire_file);
         add_object(Object);
         Current_Object = Object;
         }
      }
}

void
display_title_string(char *title)
{
   int pts[8];
   /* Clear out the old text */
#ifdef TURBOC
   setfillstyle(SOLID_FILL, 16);
   pts[0] = 0;              pts[1] = 10;
   pts[2] = window_width-1; pts[3] = 10;
   pts[4] = window_width-1; pts[5] = 20;
   pts[6] = 0;              pts[7] = 20;
   fillpoly(4, &pts[0]);
   setcolor(10);
   outtextxy(2, 10, title);
#else
   printf("%s", title);
#endif
}

void
show_orientation()
{
   char *mod_text;
   int pts[8];

   /* Clear out the old text */
#ifdef TURBOC
   setfillstyle(SOLID_FILL, 16);
   pts[0] = 0;              pts[1] = 0;
   pts[2] = window_width-1; pts[3] = 0;
   pts[4] = window_width-1; pts[5] = 20;
   pts[6] = 0;              pts[7] = 20;
   fillpoly(4, &pts[0]);
#endif

   /* Figure out the option string */
   if (Options & Rotate_Option1)
      mod_text = "(Rotate view position, keeping the focal point) ";
   else if (Options & Rotate_Option2)
      mod_text = "(Rotate the current object about the origin)    ";
   else if (Options & Rotate_Option3)
      mod_text = "(Rotate the current object about its own center)";
   else if (Options & Translate_Option1)
      mod_text = "(Translate the view position)                   ";
   else if (Options & Translate_Option2)
      mod_text = "(Translate the current object position)         ";
   else
      mod_text = "(unknown option)           ";

   /* Draw in the new text */
#ifdef TURBOC
   sprintf(title, "%s", mod_text);
   setcolor(10);
   outtextxy(2, 0, title);
#else
   printf("%s\n", mod_text);
#endif
}

int
do_command(ch)
   char ch;
{
   OBJECT *Temp_Object;
   VECTOR Angles, Offset;
   FILE *dump_file;
   int redraw = 0;

   switch(ch) {
   case 9:
      Current_Object = Current_Object->Next_Object;
      if (Current_Object == NULL) Current_Object = All_Objects;
      redraw = 1;
      break;
   case 27:
      reset_viewpoint();
      break;
   case 'A':
      All_Flag = (1 - All_Flag);
      redraw = 1;
      break;
   case 'b':
      if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 0.0, -1.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, 0.0, -1.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'B':
      if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 0.0, -5.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 0.0, -5.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'c':
      restorecrtmode();
      create_new_object();
      setgraphmode(getgraphmode());
      reset_pallette();
      redraw = 1;
      break;
   case 'C':
      clearviewport();
      redraw = 1;
      break;
   case 'd':
      if (Options & Rotate_Option1) {
         VSet(Angles, degrees_to_radians(-5), 0.0, 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, degrees_to_radians(5), 0.0, 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, degrees_to_radians(5), 0.0, 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 0.0, -1.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, -1.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'D':
      if (Options & Rotate_Option1) {
         VSet(Angles, degrees_to_radians(-30), 0.0, 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, degrees_to_radians(30), 0.0, 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, degrees_to_radians(30), 0.0, 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 0.0, -5.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, -5.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'f':
      if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 0.0, 1.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, 0.0, 1.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'F':
      if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 0.0, 5.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, 0.0, 5.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'G':
      screen_dump();
      break;
   case 'i':
      dump_file = fopen("dump.val", "rt");
      read_wire_data(dump_file);
      fclose(dump_file);
      redraw = 1;
      break;
   case 'j':
      Draw_Option = (1 - Draw_Option);
      redraw = 1;
      break;
   case 'J':
      Draw_Axes_Flag = (1 - Draw_Axes_Flag);
      redraw = 1;
      break;
   case 'k':
      Show_Orientation_Flag = (1 - Show_Orientation_Flag);
      redraw = 1;
      break;
   case 'l':
      if (Options & Rotate_Option1) {
         VSet(Angles, 0.0, degrees_to_radians(5), 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, 0.0, degrees_to_radians(-5), 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, 0.0, degrees_to_radians(-5), 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, -1.0, 0.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, -1.0, 0.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'L':
      if (Options & Rotate_Option1) {
         VSet(Angles, 0.0, degrees_to_radians(30), 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, 0.0, degrees_to_radians(-30), 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, 0.0, degrees_to_radians(-30), 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, -5.0, 0.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, -5.0, 0.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'm':
      edit_lights();
      redraw = 1;
      break;
   case 'M':
      if (Options & (Translate_Option1 | Rotate_Option1)) {
         restorecrtmode();
         printf("Modify viewpoint\n\n");
         modify_object((OBJECT *)Eye);
         Normalize_View(&WS);
         setgraphmode(getgraphmode());
         reset_pallette();
         }
      else {
         restorecrtmode();
         printf("Modify object\n\n");
         modify_object(Current_Object);
         setgraphmode(getgraphmode());
         reset_pallette();
         }
      redraw = 1;
      break;
   case 'n':
      Current_Object = Current_Object->Next_Object;
      if (Current_Object == NULL) Current_Object = All_Objects;
      redraw = 1;
      break;
   case 'O':
      dump_wire_data();
      redraw = 1;
      break;
   case 'o':
      prepare_DKB_output();
      redraw = 1;
      break;
   case 'p':
      if (Options & Translate_Option1)
         Options = Rotate_Option1;
      else if (Options & Rotate_Option1)
         Options = Translate_Option2;
      else if (Options & Translate_Option2)
         Options = Rotate_Option2;
      else if (Options & Rotate_Option2)
         Options = Rotate_Option3;
      else if (Options & Rotate_Option3)
         Options = Translate_Option1;
      else
         Options = Rotate_Option1;
      show_orientation();
      break;
   case 'P':
      if (Options & Translate_Option1)
         Options = Rotate_Option3;
      else if (Options & Rotate_Option1)
         Options = Translate_Option1;
      else if (Options & Translate_Option2)
         Options = Rotate_Option1;
      else if (Options & Rotate_Option2)
         Options = Translate_Option2;
      else if (Options & Rotate_Option3)
         Options = Rotate_Option2;
      else
         Options = Rotate_Option1;
      show_orientation();
      break;
   case 'q':
   case 'Q':
      closegraph();
      exit(0);
   case 'r':
      if (Options & Rotate_Option1) {
         VSet(Angles, 0.0, degrees_to_radians(-5), 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, 0.0, degrees_to_radians(5), 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, 0.0, degrees_to_radians(5), 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 1.0, 0.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 1.0, 0.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'R':
      if (Options & Rotate_Option1) {
         VSet(Angles, 0.0, degrees_to_radians(-30), 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, 0.0, degrees_to_radians(30), 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, 0.0, degrees_to_radians(30), 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 5.0, 0.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 5.0, 0.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 's':
      if (All_Objects != NULL) {
         if (All_Flag)
            render_objects();
         else
            render_current_object();
         }
      break;
   case 't':
      /* Dump normal triangles of all objects */
      prepare_triangle_output(0);
      break;
   case 'T':
      /* Dump normal triangles of all objects */
      prepare_triangle_output(1);
      break;
   case 'u':
      if (Options & Rotate_Option1) {
         VSet(Angles, degrees_to_radians(5), 0.0, 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, degrees_to_radians(-5), 0.0, 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, degrees_to_radians(-5), 0.0, 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 1.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, 1.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'U':
      if (Options & Rotate_Option1) {
         VSet(Angles, degrees_to_radians(30), 0.0, 0.0);
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, degrees_to_radians(-30), 0.0, 0.0);
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, degrees_to_radians(-30), 0.0, 0.0);
         Rotate2(Current_Object, &Angles);
         }
      else if (Options & Translate_Option1) {
         VSet(Offset, 0.0, 5.0, 0.0);
         Translate((OBJECT *)Eye, &Offset);
         }
      else if (Options & Translate_Option2) {
         VSet(Offset, 0.0, 5.0, 0.0);
         Translate(Current_Object, &Offset);
         }
      redraw = 1;
      break;
   case 'x':
      delete_object(Current_Object);
      redraw = 1;
      break;
   case 'z':
      if (Options & Rotate_Option1) {
         VSet(Angles, 0.0, 0.0, degrees_to_radians(5));
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, 0.0, 0.0, degrees_to_radians(-5));
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, 0.0, 0.0, degrees_to_radians(-5));
         Rotate2(Current_Object, &Angles);
         }
      redraw = 1;
      break;
   case 'Z':
      if (Options & Rotate_Option1) {
         VSet(Angles, 0.0, 0.0, degrees_to_radians(-5));
         Rotate2((OBJECT *)Eye, &Angles);
         }
      else if (Options & Rotate_Option2) {
         VSet(Angles, 0.0, 0.0, degrees_to_radians(5));
         Rotate1(Current_Object, &Angles);
         }
      else if (Options & Rotate_Option3) {
         VSet(Angles, 0.0, 0.0, degrees_to_radians(5));
         Rotate2(Current_Object, &Angles);
         }
      redraw = 1;
      break;
   }
   return redraw;
}

void
init_globals()
{
   int i, j;

   PATTERN = 0xFFFFFFFF;
   LINEWIDTH = 1;
   Draw_Option = 0;

   Eye = (VIEWPOINT *)create_object(Viewpoint_Type);
   Eye->Focus->Aspect = (DBL)window_width / (DBL)window_height;
   Normalize_View(&WS);

   Light_Sources  = NULL;
   All_Objects    = NULL;
   Current_Object = NULL;
}

void
show_current_object()
{
   if (Current_Object != NULL)
      Show(Current_Object, &RGB_White);
}

int
allocate_map_buffers()
{
   int i, l, offset, strip_count;
   long Depth_Elements_Per_Buffer;
   long Screen_Elements_Per_Buffer;
   long total_map_elements = (long)window_width * (long)window_height;
   long buffers_needed = total_map_elements /
                         (CHUNK / max(sizeof(depth_type),
                                      sizeof(color_type))) + 1;

   Total_Map_Buffers = 0;
   for (i=0;i<buffers_needed;i++) {
      Depth_Map[i]  = (void *)malloc(CHUNK);
      Screen_Map[i] = (void *)malloc(CHUNK);
      if (Depth_Map == NULL || Screen_Map[i] == NULL) break;
      Total_Map_Buffers++;
      }
   if (Total_Map_Buffers == 0) {
      printf("Failed to allocate buffers\n");
      return 0;
      }
   Depth_Elements_Per_Buffer = CHUNK / sizeof(unsigned int);
   Screen_Elements_Per_Buffer = CHUNK / sizeof(color_type);
   Total_Lines_Per_Buffer = min(Depth_Elements_Per_Buffer,
                                Screen_Elements_Per_Buffer) / window_width;
   Total_Elements_Per_Buffer = Total_Lines_Per_Buffer * window_width;
   Total_Map_Lines = Total_Lines_Per_Buffer * Total_Map_Buffers;
   Total_Elements = Total_Elements_Per_Buffer * Total_Map_Buffers;
}

void
deallocate_map_buffers()
{
   int i;
   for (i=0;i<Total_Map_Buffers;i++) {
      free(Depth_Map[i]);
      free(Screen_Map[i]);
      }
}

void
reset_map_buffers()
{
   int i, offset;
   for (i=0;i<Total_Map_Buffers;i++)
      for (offset=0;offset<Total_Elements_Per_Buffer;offset++) {
         Depth_Map[i][offset] = 65535;
         Screen_Map[i][offset].red   = 0;
         Screen_Map[i][offset].green = 0;
         Screen_Map[i][offset].blue  = 0;
         }
}

void
render_current_object()
{
   int i, l, offset, strip_count;

   Shade_Flag = 1;
   if (!allocate_map_buffers()) return;
   setup_renderer();
   if (Current_Object != NULL) {
      strip_count = (window_height + Total_Map_Lines - 1) / Total_Map_Lines;
      for (l=0;l<strip_count;l++) {
         First_Map_Line = l * Total_Map_Lines;
         Last_Map_Line = (l + 1) * Total_Map_Lines - 1;
         setcolor(224);
         line(actual_view.left, First_Map_Line,
              actual_view.left, Last_Map_Line);
         line(actual_view.right, First_Map_Line,
              actual_view.right, Last_Map_Line);
         reset_map_buffers();
         Show(Current_Object, &RGB_White);
         }
      }
   deallocate_map_buffers();
   Shade_Flag = 0;
}

void
show_objects()
{
   OBJECT *Object;
   for (Object = All_Objects;
        Object != NULL;
        Object = Object->Next_Object) {
      if (Object == Current_Object)
         Show(Object, &RGB_White)
      else
         Show(Object, &RGB_Red)
      }
}

void
render_objects()
{
   OBJECT *Object;
   int l, offset, strip_count, i, j, pts[8];
   if (!allocate_map_buffers()) return;
   TargaOpen("data.tga", window_width, window_height);
   Shade_Flag = 1;
   strip_count = (window_height + Total_Map_Lines - 1) / Total_Map_Lines;
   for (l=0;l<strip_count;l++) {
      First_Map_Line = l * Total_Map_Lines;
      Last_Map_Line = (l + 1) * Total_Map_Lines - 1;
      setcolor(224);
      line(actual_view.left, First_Map_Line,
           actual_view.left, Last_Map_Line);
      line(actual_view.right, First_Map_Line,
           actual_view.right, Last_Map_Line);
      reset_map_buffers();
      for (Object = All_Objects;
           Object != NULL;
           Object = Object->Next_Object) {
         if (Object == Current_Object)
            Show(Object, &RGB_White)
         else
            Show(Object, &RGB_Red)
         }
      TargaWrite();
      }
   TargaClose();
   deallocate_map_buffers();
   Shade_Flag = 0;
}

void
screen_dump()
{
   TargaOpen("dump.tga", window_width, window_height);
   IndexedWrite();
   TargaClose();
}

int
main(argc, argv)
   int argc;
   char **argv;
{
   char ch = 0;
   FILE *read_file;

#ifdef TURBOC
#else
/* For Mac
   argc = ccommand(&argv);
*/
   console_options.top  = 340;
   console_options.left = 30;
   console_options.nrows = 10;
   console_options.ncols = 80;
   console_options.pause_atexit = 0;
   cshow(stdout);
   csetmode(C_RAW, stdin);
   argc = ccommand(&argv);
#endif

   init_signal_catcher();
   initialize_graphics();
   init_globals();
   clearviewport();
   setup_renderer();
   if (argc == 2) {
      read_file = fopen(argv[1], "rt");
      if (read_file != NULL)
      read_wire_data(read_file);
      Normalize_View(&WS);
      fclose(read_file);
      }
   drawRect(actual_view.left, actual_view.top,
            actual_view.right, actual_view.bottom, BLUE);
   show_orientation();
   if (All_Objects != NULL) show_current_object();

   for (;;) {
      ch = getch();
      if (do_command(ch)) {
         clearviewport();
         drawRect(actual_view.left, actual_view.top,
                  actual_view.right, actual_view.bottom, BLUE);

         if (Draw_Axes_Flag)
            drawaxes();

         if (Show_Orientation_Flag)
            show_orientation();

         if (All_Objects != NULL) {
            if (All_Flag)
               show_objects();
            else
               show_current_object();
            }
         }
      }
   return 0;
}
